from unittest import TestCase

from lxml import etree

from pcs.common.str_tools import format_list

from pcs_test.tier0.lib.commands.test_cluster_property import ALLOWED_PROPERTIES
from pcs_test.tools.assertions import ac
from pcs_test.tools.cib import (
    get_assert_pcs_effect_mixin_old as get_assert_pcs_effect_mixin,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunnerOld as PcsRunner
from pcs_test.tools.pcs_runner import pcs_old as pcs

# pylint: disable=invalid-name

empty_cib = rc("cib-empty.xml")


def get_invalid_option_messages(option_name, error=True):
    error_occurred = (
        "Error: Errors have occurred, therefore pcs is unable to continue\n"
    )
    use_force = ", use --force to override"
    return (
        "{severity}: invalid cluster property option '{option_name}', allowed "
        "options are: {allowed_properties}{use_force}\n"
        "{error_occurred}"
    ).format(
        severity="Error" if error else "Warning",
        option_name=option_name,
        allowed_properties=format_list(ALLOWED_PROPERTIES),
        use_force=use_force if error else "",
        error_occurred=error_occurred if error else "",
    )


class PropertyTest(TestCase):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_properties")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def testEmpty(self):
        output, returnVal = pcs(self.temp_cib.name, ["property"])
        assert returnVal == 0, "Unable to list resources"
        assert output == "Cluster Properties:\n", [output]

    def testDefaults(self):
        output, returnVal = pcs(
            self.temp_cib.name, "property --defaults".split()
        )
        prop_defaults = output
        assert returnVal == 0, "Unable to list resources"
        assert output.startswith("Cluster Properties:\n batch-limit")

        output, returnVal = pcs(self.temp_cib.name, "property --all".split())
        assert returnVal == 0, "Unable to list resources"
        assert output.startswith("Cluster Properties:\n batch-limit")
        ac(output, prop_defaults)

        output, returnVal = pcs(
            self.temp_cib.name, "property set blahblah=blah".split()
        )
        assert returnVal == 1
        assert output == get_invalid_option_messages("blahblah"), [output]

        output, returnVal = pcs(
            self.temp_cib.name, "property set blahblah=blah --force".split()
        )
        assert returnVal == 0, output
        assert output == get_invalid_option_messages(
            "blahblah", error=False
        ), output

        output, returnVal = pcs(
            self.temp_cib.name, "property set stonith-enabled=false".split()
        )
        assert returnVal == 0, output
        assert output == "", output

        output, returnVal = pcs(self.temp_cib.name, ["property"])
        assert returnVal == 0
        assert (
            output
            == "Cluster Properties:\n blahblah: blah\n stonith-enabled: false\n"
        ), [output]

        output, returnVal = pcs(
            self.temp_cib.name, "property --defaults".split()
        )
        assert returnVal == 0, "Unable to list resources"
        assert output.startswith("Cluster Properties:\n batch-limit")
        ac(output, prop_defaults)

        output, returnVal = pcs(self.temp_cib.name, "property --all".split())
        assert returnVal == 0, "Unable to list resources"
        assert "blahblah: blah" in output
        assert "stonith-enabled: false" in output
        assert output.startswith("Cluster Properties:\n batch-limit")

        output, returnVal = pcs(
            self.temp_cib.name, "property config stonith-action".split()
        )
        assert returnVal == 0, output
        assert "stonith-action: reboot" in output

    def testBadProperties(self):
        o, r = pcs(self.temp_cib.name, "property set xxxx=zzzz".split())
        self.assertEqual(r, 1)
        ac(
            o,
            get_invalid_option_messages("xxxx"),
        )
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        output, returnVal = pcs(
            self.temp_cib.name, "property set =5678 --force".split()
        )
        ac(output, "Error: missing key in '=5678' option\n")
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        output, returnVal = pcs(
            self.temp_cib.name, "property set =5678".split()
        )
        ac(output, "Error: missing key in '=5678' option\n")
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        output, returnVal = pcs(
            self.temp_cib.name, "property set bad_format".split()
        )
        ac(output, "Error: missing value of 'bad_format' option\n")
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        output, returnVal = pcs(
            self.temp_cib.name, "property set bad_format --force".split()
        )
        ac(output, "Error: missing value of 'bad_format' option\n")
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        o, r = pcs(self.temp_cib.name, "property unset zzzzz".split())
        self.assertEqual(r, 1)
        ac(
            o,
            "Error: Cannot remove property 'zzzzz', it is not present in "
            "property set 'cib-bootstrap-options', use --force to override\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n",
        )
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

        o, r = pcs(self.temp_cib.name, "property unset zzzz --force".split())
        self.assertEqual(r, 0)
        ac(
            o,
            "Warning: Cannot remove property 'zzzz', it is not present in "
            "property set 'cib-bootstrap-options'\n",
        )
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(o, "Cluster Properties:\n")

    def test_set_property_validation_enum(self):
        output, returnVal = pcs(
            self.temp_cib.name, "property set no-quorum-policy=freeze".split()
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 no-quorum-policy: freeze
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set no-quorum-policy=freeze --force".split(),
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 no-quorum-policy: freeze
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set no-quorum-policy=not_valid_value".split(),
        )
        ac(
            output,
            (
                "Error: 'not_valid_value' is not a valid no-quorum-policy "
                "value, use 'demote', 'freeze', 'ignore', 'stop', 'suicide', "
                "use --force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 no-quorum-policy: freeze
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set no-quorum-policy=not_valid_value --force".split(),
        )
        ac(
            output,
            (
                "Warning: 'not_valid_value' is not a valid no-quorum-policy "
                "value, use 'demote', 'freeze', 'ignore', 'stop', 'suicide'\n"
            ),
        )
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 no-quorum-policy: not_valid_value
""",
        )

    def test_set_property_validation_boolean(self):
        output, returnVal = pcs(
            self.temp_cib.name, "property set enable-acl=TRUE".split()
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 enable-acl: TRUE
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name, "property set enable-acl=no".split()
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 enable-acl: no
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set enable-acl=TRUE --force".split(),
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 enable-acl: TRUE
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set enable-acl=not_valid_value".split(),
        )
        ac(
            output,
            (
                "Error: 'not_valid_value' is not a valid enable-acl value, use "
                "a pacemaker boolean value: '0', '1', 'false', 'n', 'no', "
                "'off', 'on', 'true', 'y', 'yes', use --force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 enable-acl: TRUE
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set enable-acl=not_valid_value --force".split(),
        )
        ac(
            output,
            (
                "Warning: 'not_valid_value' is not a valid enable-acl value, "
                "use a pacemaker boolean value: '0', '1', 'false', 'n', 'no', "
                "'off', 'on', 'true', 'y', 'yes'\n"
            ),
        )
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 enable-acl: not_valid_value
""",
        )

    def test_set_property_validation_integer(self):
        output, returnVal = pcs(
            self.temp_cib.name, "property set migration-limit=0".split()
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 migration-limit: 0
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name, "property set migration-limit=-10".split()
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 migration-limit: -10
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set migration-limit=0 --force".split(),
        )
        ac(output, "")
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 migration-limit: 0
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name, "property set migration-limit=0.1".split()
        )
        ac(
            output,
            (
                "Error: '0.1' is not a valid migration-limit value, use an "
                "integer or INFINITY or -INFINITY, use --force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assertEqual(returnVal, 1)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 migration-limit: 0
""",
        )

        output, returnVal = pcs(
            self.temp_cib.name,
            "property set migration-limit=0.1 --force".split(),
        )
        ac(
            output,
            (
                "Warning: '0.1' is not a valid migration-limit value, use an "
                "integer or INFINITY or -INFINITY\n"
            ),
        )
        self.assertEqual(returnVal, 0)
        o, _ = pcs(self.temp_cib.name, "property config".split())
        ac(
            o,
            """Cluster Properties:
 migration-limit: 0.1
""",
        )


class PropertyUnset(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//crm_config")[0]
        )
    ),
):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_properties_unset")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    @staticmethod
    def fixture_xml_no_props():
        # must match empty_cib
        return """
            <crm_config />
        """

    @staticmethod
    def fixture_xml_empty_props():
        # must match empty_cib
        return """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options" />
            </crm_config>
        """

    @staticmethod
    def fixture_xml_with_props():
        # must match empty_cib
        return """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-batch-limit"
                        name="batch-limit" value="100"
                    />
                </cluster_property_set>
            </crm_config>
        """

    def test_keep_empty_nvset(self):
        self.assert_effect(
            "property set batch-limit=100".split(),
            self.fixture_xml_with_props(),
        )
        self.assert_effect(
            "property unset batch-limit".split(), self.fixture_xml_empty_props()
        )

    def test_dont_create_nvset_on_removal(self):
        self.assert_pcs_fail(
            "property unset batch-limit".split(),
            (
                "Error: Cannot remove property 'batch-limit', it is not present"
                " in property set 'cib-bootstrap-options', use --force to "
                "override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation


def _get_primitive_fixture(
    res_id, agent_standard="ocf", agent_provider="heartbeat", agent_type="Dummy"
):
    _provider = ""
    if agent_provider:
        _provider = f' provider="{agent_provider}"'
    return f"""
        <primitive class="{agent_standard}" id="{res_id}"{_provider} type="{agent_type}">
            <operations>
                <op id="{res_id}-monitor-interval-10s" interval="10s"
                    name="monitor" timeout="20s"/>
            </operations>
        </primitive>
    """


FIXTURE_DUMMY = _get_primitive_fixture("Dummy")
FIXTURE_PRIMITIVE_FOR_CLONE = _get_primitive_fixture("C")

FIXTURE_CLONE = f"""<clone id="C-clone">{FIXTURE_PRIMITIVE_FOR_CLONE}</clone>"""

FIXTURE_STONITH_FOR_CLONE = """
    <primitive class="stonith" id="fence-device" type="fence_xvm">
        <operations>
          <op id="fence-device-monitor-interval-60s" interval="60s"
              name="monitor"/>
        </operations>
    </primitive>
"""

FIXTURE_STONITH_CLONE = f"""
    <clone id="fence-device-clone">{FIXTURE_STONITH_FOR_CLONE}</clone>
"""

FIXTURE_STONITH_PROMOTABLE = f"""
    <clone id="fence-device-clone">
        {FIXTURE_STONITH_FOR_CLONE}
        <meta_attributes id="fence-device-clone-meta_attributes">
            <nvpair id="fence-device-clone-meta_attributes-promotable"
                name="promotable" value="true"></nvpair>
        </meta_attributes>
    </clone>
"""
FIXTURE_CLONE_WITH_OPTIONS = f"""
    <clone id="CustomCloneId">
        {FIXTURE_PRIMITIVE_FOR_CLONE}
        <meta_attributes id="CustomCloneId-meta_attributes">
            <nvpair id="CustomCloneId-meta_attributes-a" name="a" value="b"/>
            <nvpair id="CustomCloneId-meta_attributes-c" name="c" value="d"/>
            <nvpair id="CustomCloneId-meta_attributes-globally-unique"
                name="globally-unique" value="true"/>
        </meta_attributes>
    </clone>
"""

FIXTURE_CLONED_GROUP = """
    <clone id="Group-clone">
        <group id="Group">
            <primitive class="ocf" id="G1" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="G1-monitor-interval-10s" interval="10s"
                        name="monitor" timeout="20s"/>
                </operations>
            </primitive>
            <primitive class="ocf" id="G2" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="G2-monitor-interval-10s" interval="10s"
                        name="monitor" timeout="20s"/>
                </operations>
            </primitive>
        </group>
    </clone>
"""

FIXTURE_GROUP_WITH_STONITH = f"""
    <group id="Group">{FIXTURE_STONITH_FOR_CLONE}</group>
"""

FIXTURE_CLONED_GROUP_WITH_STONITH = f"""
    <clone id="Group-clone">{FIXTURE_GROUP_WITH_STONITH}</clone>
"""

FIXTURE_GROUP_LAST_MEMBER = f"""
    <group id="Group">{FIXTURE_PRIMITIVE_FOR_CLONE}</group>
"""

FIXTURE_CONSTRAINTS_CONFIG_XML = """
    <constraints>
        <rsc_location id="location-C-clone-rh7-1-INFINITY" node="rh7-1"
            rsc="C-clone" score="INFINITY"/>
        <rsc_location id="location-TagCloneOnly-rh7-1-INFINITY"
            node="rh7-1" rsc="TagCloneOnly" score="INFINITY"/>
    </constraints>
"""

FIXTURE_TAGS_CONFIG_XML = """
    <tags>
        <tag id="TagCloneOnly">
            <obj_ref id="C-clone"/>
        </tag>
        <tag id="TagNotCloneOnly">
            <obj_ref id="C-clone"/>
            <obj_ref id="Dummy"/>
        </tag>
    </tags>
"""

FIXTURE_TAGS_RESULT_XML = """
    <tags>
        <tag id="TagNotCloneOnly">
            <obj_ref id="Dummy"/>
        </tag>
    </tags>
"""

FIXTURE_BUNDLE_RESOURCE = """
    <bundle id="Dummy-bundle">
        <docker image="pcs:test" />
        {}
    </bundle>
""".format(
    FIXTURE_DUMMY
)


def fixture_resources_xml(*resources_xml_list):
    return """
        <resources>
            {0}
        </resources>
    """.format(
        "\n".join(resources_xml_list)
    )


def fixture_clone(clone_id, primitive_id, promotable=False):
    parts = []
    parts.append(f"""<clone id="{clone_id}">""")
    parts.append(
        f"""
        <primitive class="ocf" id="{primitive_id}" provider="heartbeat"
            type="Dummy">
            <operations>
                <op id="{primitive_id}-monitor-interval-10s" interval="10s"
                    name="monitor" timeout="20s"/>
            </operations>
        </primitive>
        """
    )
    if promotable:
        parts.append(
            f"""
            <meta_attributes id="{clone_id}-meta_attributes">
                <nvpair id="{clone_id}-meta_attributes-promotable"
                    name="promotable" value="true">
                </nvpair>
            </meta_attributes>
            """
        )
    parts.append("""</clone>""")
    return "".join(parts)


FIXTURE_CLONE_AND_RESOURCE = fixture_resources_xml(
    fixture_clone("C-clone", "C"), FIXTURE_DUMMY
)


FIXTURE_RESOURCES = fixture_resources_xml(
    FIXTURE_DUMMY, FIXTURE_PRIMITIVE_FOR_CLONE
)


def fixture_clone_stonith_msg(forced=False, group=False):
    return (
        "{severity}: {group}No need to clone stonith resource 'fence-device', any node"
        " can use a stonith resource (unless specifically banned) regardless of "
        "whether the stonith resource is running on that node or not{use_force}"
    ).format(
        severity="Warning" if forced else "Error",
        group="Group 'Group' contains stonith resource. " if group else "",
        use_force="\n" if forced else ", use " "--force to override\n",
    )


class Unclone(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    empty_cib = rc("cib-empty.xml")

    def assert_tags_xml(self, expected_xml):
        self.assert_resources_xml_in_cib(
            expected_xml,
            get_cib_part_func=lambda cib: etree.tostring(
                etree.parse(cib).findall(".//tags")[0],
            ),
        )

    def assert_constraint_xml(self, expected_xml):
        self.assert_resources_xml_in_cib(
            expected_xml,
            get_cib_part_func=lambda cib: etree.tostring(
                etree.parse(cib).findall(".//constraints")[0],
            ),
        )

    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_cib_resource_group_ungroup")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        xml_manip = XmlManipulation.from_file(self.empty_cib)
        xml_manip.append_to_first_tag_name(
            "resources",
            FIXTURE_CLONE,
            FIXTURE_DUMMY,
        )
        xml_manip.append_to_first_tag_name(
            "configuration",
            FIXTURE_TAGS_CONFIG_XML,
        )
        xml_manip.append_to_first_tag_name(
            "constraints",
            """
            <rsc_location id="location-C-clone-rh7-1-INFINITY" node="rh7-1"
                rsc="C-clone" score="INFINITY"/>
            """,
            """
            <rsc_location id="location-TagCloneOnly-rh7-1-INFINITY"
                node="rh7-1" rsc="TagCloneOnly" score="INFINITY"/>
            """,
        )
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()

    def test_nonexistent_clone(self):
        self.assert_pcs_fail(
            "resource unclone NonExistentClone".split(),
            "Error: could not find resource: NonExistentClone\n",
        )
        self.assert_resources_xml_in_cib(FIXTURE_CLONE_AND_RESOURCE)
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CONSTRAINTS_CONFIG_XML)

    def test_not_clone_resource(self):
        self.assert_pcs_fail(
            "resource unclone Dummy".split(),
            "Error: 'Dummy' is not a clone resource\n",
        )
        self.assert_resources_xml_in_cib(FIXTURE_CLONE_AND_RESOURCE)
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CONSTRAINTS_CONFIG_XML)

    def test_unclone_clone_id(self):
        self.assert_effect(
            "resource unclone C-clone".split(), FIXTURE_RESOURCES
        )
        self.assert_tags_xml(FIXTURE_TAGS_RESULT_XML)
        self.assert_constraint_xml("<constraints/>")

    def test_unclone_resource_id(self):
        self.assert_effect("resource unclone C".split(), FIXTURE_RESOURCES)
        self.assert_tags_xml(FIXTURE_TAGS_RESULT_XML)
        self.assert_constraint_xml("<constraints/>")


class Clone(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    # pylint: disable=too-many-public-methods
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cib_resource_clone_unclone_clone")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.set_cib_file(FIXTURE_PRIMITIVE_FOR_CLONE)
        self.stonith_deprecation_warning = (
            "Deprecation Warning: Ability of this command to accept stonith "
            "resources is deprecated and will be removed in a future release.\n"
        )

    def tearDown(self):
        self.temp_cib.close()

    def set_cib_file(self, *xml_string_list):
        xml_manip = XmlManipulation.from_file(self.empty_cib)
        xml_manip.append_to_first_tag_name("resources", *xml_string_list)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def test_clone(self):
        self.assert_effect(
            "resource clone C".split(),
            fixture_resources_xml(fixture_clone("C-clone", "C")),
        )

    def test_clone_custom_id(self):
        self.assert_effect(
            "resource clone C CustomCloneId".split(),
            fixture_resources_xml(fixture_clone("CustomCloneId", "C")),
        )

    def test_clone_id_increment(self):
        self.set_cib_file(
            fixture_clone("C-clone", "Dummy"),
            FIXTURE_PRIMITIVE_FOR_CLONE,
        )
        self.assert_effect(
            "resource clone C".split(),
            fixture_resources_xml(
                fixture_clone("C-clone", "Dummy"),
                fixture_clone("C-clone-1", "C"),
            ),
        )

    def test_clone_id_is_stonith(self):
        self.set_cib_file(FIXTURE_STONITH_FOR_CLONE)
        self.assert_pcs_fail(
            "resource clone fence-device".split(),
            self.stonith_deprecation_warning + fixture_clone_stonith_msg(),
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_STONITH_FOR_CLONE)
        )

    def test_clone_id_is_stonith_forced(self):
        self.set_cib_file(FIXTURE_STONITH_FOR_CLONE)
        self.assert_effect(
            "resource clone fence-device --force".split(),
            fixture_resources_xml(FIXTURE_STONITH_CLONE),
            stderr_full=(
                self.stonith_deprecation_warning
                + fixture_clone_stonith_msg(forced=True)
            ),
        )

    def test_clone_group_with_stonith(self):
        self.set_cib_file(FIXTURE_GROUP_WITH_STONITH)
        self.assert_effect(
            "resource clone Group".split(),
            fixture_resources_xml(FIXTURE_CLONED_GROUP_WITH_STONITH),
        )

    def test_clone_group_with_stonith_forced(self):
        self.set_cib_file(FIXTURE_GROUP_WITH_STONITH)
        self.assert_effect(
            "resource clone Group --force".split(),
            fixture_resources_xml(FIXTURE_CLONED_GROUP_WITH_STONITH),
        )

    def test_clone_globally_unique_not_ocf_agent(self):
        self.set_cib_file(
            _get_primitive_fixture(
                "C",
                agent_standard="systemd",
                agent_provider=None,
                agent_type="pacemaker",
            )
        )
        self.assert_pcs_fail(
            "resource clone C globally-unique=true".split(),
            (
                "Error: Clone option 'globally-unique' is not compatible with "
                "'systemd:pacemaker' resource agent of resource 'C'\n"
            ),
        )

    def test_clone_promotable_group_some_unsupported(self):
        self.set_cib_file(
            '<group id="G">'
            + "\n".join(
                [
                    _get_primitive_fixture(
                        "A",
                        agent_standard="systemd",
                        agent_provider=None,
                        agent_type="pacemaker",
                    ),
                    _get_primitive_fixture(
                        "B",
                        agent_provider="pacemaker",
                        agent_type="Stateful",
                    ),
                    _get_primitive_fixture(
                        "C",
                        agent_provider="pacemaker",
                    ),
                ]
            )
            + "</group>"
        )
        self.assert_pcs_fail(
            "resource clone G promotable=true".split(),
            (
                "Error: Clone option 'promotable' is not compatible with "
                "'systemd:pacemaker' resource agent of resource 'A' in group "
                "'G'\nError: Clone option 'promotable' is not compatible with "
                "'ocf:pacemaker:Dummy' resource agent of resource 'C' in group "
                "'G', use --force to override\n"
            ),
        )

    def test_clone_promotable_not_ocf_agent(self):
        self.set_cib_file(
            _get_primitive_fixture(
                "C",
                agent_standard="systemd",
                agent_provider=None,
                agent_type="pacemaker",
            )
        )
        self.assert_pcs_fail(
            "resource clone C promotable=true".split(),
            (
                "Error: Clone option 'promotable' is not compatible with "
                "'systemd:pacemaker' resource agent of resource 'C'\n"
            ),
        )

    def test_promotable_clone(self):
        self.assert_effect(
            "resource promotable C".split(),
            fixture_resources_xml(
                fixture_clone("C-clone", "C", promotable=True)
            ),
        )

    def test_promotable_clone_not_ocf_agent(self):
        self.set_cib_file(
            _get_primitive_fixture(
                "C",
                agent_standard="systemd",
                agent_provider=None,
                agent_type="pacemaker",
            )
        )
        self.assert_pcs_fail(
            "resource promotable C".split(),
            (
                "Error: Clone option 'promotable' is not compatible with "
                "'systemd:pacemaker' resource agent of resource 'C'\n"
            ),
        )

    def test_promotable_clone_unsupported_agent(self):
        self.set_cib_file(
            _get_primitive_fixture("C", agent_provider="pacemaker")
        )
        self.assert_pcs_fail(
            "resource promotable C".split(),
            (
                "Error: Clone option 'promotable' is not compatible with "
                "'ocf:pacemaker:Dummy' resource agent of resource 'C', use "
                "--force to override\n"
            ),
        )

    def test_promotable_clone_custom_id(self):
        self.assert_effect(
            "resource promotable C CustomPromotableId".split(),
            fixture_resources_xml(
                fixture_clone("CustomPromotableId", "C", promotable=True)
            ),
        )

    def test_promotable_clone_id_is_stonith(self):
        self.set_cib_file(FIXTURE_STONITH_FOR_CLONE)
        self.assert_pcs_fail(
            "resource promotable fence-device".split(),
            self.stonith_deprecation_warning + fixture_clone_stonith_msg(),
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_STONITH_FOR_CLONE)
        )

    def test_promotable_clone_id_is_stonith_forced(self):
        self.set_cib_file(FIXTURE_STONITH_FOR_CLONE)
        self.assert_pcs_fail(
            "resource promotable fence-device --force".split(),
            stderr_start=(
                self.stonith_deprecation_warning
                + fixture_clone_stonith_msg(forced=True)
            ),
        )

    def test_promotable_keyword_and_option(self):
        self.assert_pcs_fail(
            "resource promotable C CustomCloneId promotable=false".split(),
            (
                "Error: you cannot specify both promotable option and "
                "promotable keyword\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_PRIMITIVE_FOR_CLONE)
        )

    def test_clone_with_options(self):
        self.assert_effect(
            (
                "resource clone C CustomCloneId globally-unique=true meta a=b "
                "c=d"
            ).split(),
            fixture_resources_xml(FIXTURE_CLONE_WITH_OPTIONS),
        )

    def test_group_last_member(self):
        self.set_cib_file(FIXTURE_GROUP_LAST_MEMBER)
        self.assert_effect(
            "resource clone C".split(),
            fixture_resources_xml(fixture_clone("C-clone", "C")),
        )

    def test_nonexistent_resource(self):
        self.assert_pcs_fail(
            "resource clone NonExistentClone".split(),
            "Error: unable to find group or resource: NonExistentClone\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_PRIMITIVE_FOR_CLONE)
        )

    def test_invalid_clone_id(self):
        self.assert_pcs_fail(
            "resource clone C 1invalid".split(),
            "Error: invalid id '1invalid'\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_PRIMITIVE_FOR_CLONE)
        )

    def test_clone_id_already_exist(self):
        self.assert_pcs_fail(
            "resource clone C C".split(),
            "Error: id 'C' already exists\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_PRIMITIVE_FOR_CLONE)
        )

    def test_group_already_cloned(self):
        self.set_cib_file(FIXTURE_CLONED_GROUP)
        self.assert_pcs_fail(
            "resource clone Group".split(),
            "Error: cannot clone a group that has already been cloned\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_CLONED_GROUP)
        )

    def test_already_a_clone_resource(self):
        self.set_cib_file(FIXTURE_CLONED_GROUP)
        self.assert_pcs_fail(
            "resource clone G1".split(),
            "Error: G1 is already a clone resource\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_CLONED_GROUP)
        )

    def test_bundle_resource(self):
        self.set_cib_file(FIXTURE_BUNDLE_RESOURCE)
        self.assert_pcs_fail(
            "resource clone Dummy".split(),
            "Error: cannot clone bundle resource\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(FIXTURE_BUNDLE_RESOURCE)
        )



from unittest import TestCase

from pcs.cli.cluster import command


class ParseNodeAddRemote(TestCase):
    # pylint: disable=protected-access
    def test_deal_with_explicit_address(self):
        self.assertEqual(
            command._node_add_remote_separate_name_and_addr(
                ["name", "address", "a=b"]
            ),
            ("name", "address", ["a=b"]),
        )

    def test_deal_with_implicit_address(self):
        self.assertEqual(
            command._node_add_remote_separate_name_and_addr(["name", "a=b"]),
            ("name", None, ["a=b"]),
        )



import socket
from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common import reports
from pcs.common.corosync_conf import (
    CorosyncNodeAddressDto,
    CorosyncNodeAddressType,
    CorosyncNodeDto,
)
from pcs.common.interface.dto import ImplementsToDto
from pcs.lib.errors import LibraryError
from pcs.lib.validate import (
    is_ipv4_address,
    is_ipv6_address,
)


@dataclass(frozen=True)
class CorosyncNodeAddress(ImplementsToDto):
    """An address of a node loaded from corosync.conf"""

    addr: str
    link: str

    @property
    def type(self) -> CorosyncNodeAddressType:
        return get_address_type(self.addr, resolve=False)

    def to_dto(self) -> CorosyncNodeAddressDto:
        return CorosyncNodeAddressDto(
            addr=self.addr,
            link=self.link,
            type=self.type,
        )


@dataclass(frozen=True)
class CorosyncNode(ImplementsToDto):
    """A node loaded from corosync.conf"""

    name: Optional[str]
    addrs: Sequence[CorosyncNodeAddress]
    nodeid: Optional[str]

    def addr_plain_for_link(self, link: str) -> Optional[str]:
        for addr in self.addrs:
            if addr.link == link:
                return addr.addr
        return None

    def addrs_plain(self, except_link: Optional[str] = None) -> list[str]:
        return [
            addr.addr
            for addr in self.addrs
            if except_link is None or addr.link != except_link
        ]

    def to_dto(self) -> CorosyncNodeDto:
        if not self.name:
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.CorosyncConfigMissingNamesOfNodes(
                        fatal=True
                    )
                )
            )
        if not self.nodeid:
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.CorosyncConfigMissingIdsOfNodes()
                )
            )
        return CorosyncNodeDto(
            name=self.name,
            nodeid=self.nodeid,
            addrs=[addr.to_dto() for addr in self.addrs],
        )


def get_address_type(
    address: str, resolve: bool = False
) -> CorosyncNodeAddressType:
    if is_ipv4_address(address):
        return CorosyncNodeAddressType.IPV4
    if is_ipv6_address(address):
        return CorosyncNodeAddressType.IPV6
    if resolve:
        try:
            socket.getaddrinfo(address, None)
        except socket.gaierror:
            return CorosyncNodeAddressType.UNRESOLVABLE
    return CorosyncNodeAddressType.FQDN



import os.path
import signal
import subprocess

from pcs_test import TEST_ROOT

__pcs_location = os.path.join(TEST_ROOT, "pcs_for_tests")
# this can be changed from suite.py
test_installed = False


class PcsRunner:
    def __init__(self, cib_file, corosync_conf_opt=None, mock_settings=None):
        self.cib_file = cib_file
        self.corosync_conf_opt = corosync_conf_opt
        self.mock_settings = mock_settings

    def run(self, args):
        return pcs(
            self.cib_file,
            args,
            corosync_conf_opt=self.corosync_conf_opt,
            mock_settings=self.mock_settings,
        )


class PcsRunnerOld(PcsRunner):
    # TODO remove this class
    def run(self, args, ignore_stderr=False):
        stdout, stderr, retval = super().run(args)
        if ignore_stderr:
            stderr = None
        return "".join(filter(None, [stderr, stdout])), retval


def pcs(cib_file, args, corosync_conf_opt=None, mock_settings=None):
    if mock_settings is None:
        mock_settings = {}

    env_mock_settings_prefix = "PCS.SETTINGS."
    env = {
        "{}{}".format(env_mock_settings_prefix, option): value
        for option, value in mock_settings.items()
    }
    if test_installed:
        env["PCS_TEST.TEST_INSTALLED"] = "1"

    cmd = [__pcs_location]
    if cib_file:
        cmd.extend(["-f", cib_file])
    if corosync_conf_opt:
        cmd.extend(["--corosync_conf", corosync_conf_opt])
    cmd += args

    return _run(cmd, env_extend=env)


def pcs_old(
    cib_file,
    args,
    corosync_conf_opt=None,
    mock_settings=None,
    ignore_stderr=False,
):
    # TODO remove this function
    stdout, stderr, retval = pcs(
        cib_file,
        args,
        corosync_conf_opt=corosync_conf_opt,
        mock_settings=mock_settings,
    )
    if ignore_stderr:
        stderr = None
    return "".join(filter(None, [stderr, stdout])), retval


def _run(args, env_extend=None):
    env_vars = {"LC_ALL": "C"}
    env_vars.update(dict(env_extend) if env_extend else {})

    # pylint: disable=subprocess-popen-preexec-fn
    with subprocess.Popen(
        args,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        preexec_fn=(lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL)),
        close_fds=True,
        shell=False,
        env=env_vars,
        # decodes newlines and in python3 also converts bytes to str
        universal_newlines=True,
    ) as process:
        stdout, stderr = process.communicate()
        retval = process.returncode

    return stdout, stderr, retval



from pcs import settings
from pcs.common.host import (
    Destination,
    PcsKnownHost,
)

from pcs_test.tools.command_env.mock_push_cib import Call as PushCibCall
from pcs_test.tools.command_env.mock_push_corosync_conf import (
    Call as PushCorosyncConfCall,
)
from pcs_test.tools.fixture_cib import modify_cib


class EnvConfig:
    def __init__(self, call_collection):
        self.__calls = call_collection
        self.__cib_data = None
        self.__cib_tempfile = None
        self.__corosync_conf_data = None
        self.__booth = None
        self.__known_hosts_getter = None

    def set_cib_data(self, cib_data, cib_tempfile="/fake/tmp/file"):
        self.__cib_data = cib_data
        self.__cib_tempfile = cib_tempfile

    @property
    def cib_data(self):
        return self.__cib_data

    @property
    def cib_tempfile(self):
        return self.__cib_tempfile

    def set_booth(self, booth):
        self.__booth = booth

    @property
    def booth(self):
        return self.__booth

    def set_corosync_conf_data(self, corosync_conf_data):
        self.__corosync_conf_data = corosync_conf_data

    @property
    def corosync_conf_data(self):
        return self.__corosync_conf_data

    def set_known_nodes(self, host_name_list):
        """
        Set known hosts so that each host's address equals to the host's name
        list host_name_list -- list of host names
        """
        self.__known_hosts_getter = lambda: {
            name: PcsKnownHost(
                name,
                token=None,
                dest_list=[Destination(name, settings.pcsd_default_port)],
            )
            for name in host_name_list
        }

    def set_known_hosts_dests(self, know_hosts_dests):
        """
        Set known hosts so for each host a name and a dest_list is specified
        dict know_hosts_dests -- key: host name, value: list of Destination
        """
        self.__known_hosts_getter = lambda: {
            name: PcsKnownHost(name, token=None, dest_list=dest_list)
            for name, dest_list in know_hosts_dests.items()
        }

    def set_known_hosts_getter(self, known_hosts_getter):
        """
        Set a function providing known hosts
        """
        self.__known_hosts_getter = known_hosts_getter

    @property
    def known_hosts_getter(self):
        return self.__known_hosts_getter

    def push_cib(
        self,
        modifiers=None,
        name="env.push_cib",
        load_key="runner.cib.load",
        wait=-1,
        exception=None,
        instead=None,
        **modifier_shortcuts,
    ):
        """
        Create call for pushing cib.

        string name -- key of the call
        list of callable modifiers -- every callable takes etree.Element and
            returns new etree.Element with desired modification.
        string load_key -- key of a call from which stdout can be cib taken
        int wait -- wait timeout for pacemaker idle
        Exception|None exception -- exception that should raise env.push_cib
        string instead -- key of call instead of which this new call is to be
            placed
        dict modifier_shortcuts -- a new modifier is generated from each
            modifier shortcut.
            As key there can be keys of MODIFIER_GENERATORS.
            Value is passed into appropriate generator from MODIFIER_GENERATORS.
            For details see pcs_test.tools.fixture_cib (mainly the variable
            MODIFIER_GENERATORS - please refer it when you are adding params
            here)
        """
        cib_xml = modify_cib(
            self.__calls.get(load_key).stdout, modifiers, **modifier_shortcuts
        )
        self.__calls.place(
            name,
            PushCibCall(cib_xml, wait_timeout=wait, exception=exception),
            instead=instead,
        )

    def push_cib_custom(
        self,
        name="env.push_cib_custom",
        custom_cib=None,
        wait=-1,
        exception=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            PushCibCall(
                custom_cib,
                custom_cib=True,
                wait_timeout=wait,
                exception=exception,
            ),
            instead=instead,
        )

    def push_corosync_conf(
        self,
        name="env.push_corosync_conf",
        corosync_conf_text="",
        skip_offline_targets=False,
        raises=False,
        need_stopped_cluster=False,
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            PushCorosyncConfCall(
                corosync_conf_text,
                skip_offline_targets,
                raises=raises,
                need_stopped_cluster=need_stopped_cluster,
            ),
            instead=instead,
            before=before,
        )



import unittest

from pcs_test.tools.color_text_runner.format import (
    Format,
    Output,
    separator1,
)
from pcs_test.tools.color_text_runner.writer import (
    DotWriter,
    ImprovedVerboseWriter,
    StandardVerboseWriter,
    TimeWriter,
    Writer,
)


def get_text_test_result_class(
    slash_last_fail_in_overview=False,
    traditional_verbose=False,
    traceback_highlight=False,
    fast_info=False,
    rich_format=True,
    measure_time=False,
):
    class ColorTextTestResult(unittest.TextTestResult):
        # pylint: disable=bad-super-call, invalid-name
        def __init__(self, stream, descriptions, verbosity):
            super().__init__(stream, descriptions, verbosity)
            self.verbosity = 2 if traditional_verbose else verbosity

            self._format = Format(Output(rich_format))
            self.reportWriter = self.__chooseWriter()(
                self.stream,
                self._format,
                self.descriptions,
                traceback_highlight,
                fast_info,
            )
            self.skip_map = {}

        def startTest(self, test):
            super(unittest.TextTestResult, self).startTest(test)
            self.reportWriter.startTest(test)

        def addSuccess(self, test):
            super(unittest.TextTestResult, self).addSuccess(test)
            self.reportWriter.addSuccess(test)

        def addError(self, test, err):
            super(unittest.TextTestResult, self).addError(test, err)
            self.reportWriter.addError(test, err, traceback=self.errors[-1][1])

        def addFailure(self, test, err):
            super(unittest.TextTestResult, self).addFailure(test, err)
            self.reportWriter.addFailure(
                test, err, traceback=self.failures[-1][1]
            )

        def addSkip(self, test, reason):
            super(unittest.TextTestResult, self).addSkip(test, reason)
            self.skip_map.setdefault(reason, []).append(test)
            self.reportWriter.addSkip(test, reason)

        def addExpectedFailure(self, test, err):
            super(unittest.TextTestResult, self).addExpectedFailure(test, err)
            self.reportWriter.addExpectedFailure(test, err)

        def addUnexpectedSuccess(self, test):
            super(unittest.TextTestResult, self).addUnexpectedSuccess(test)
            self.reportWriter.addUnexpectedSuccess(test)

        def printErrors(self):
            line_list = self._format.error_list(
                "ERROR",
                self.errors,
                self.descriptions,
                traceback_highlight,
            ) + self._format.error_list(
                "FAIL",
                self.failures,
                self.descriptions,
                traceback_highlight,
            )

            if (self.errors + self.failures) or self.skip_map:
                line_list.extend([separator1, ""])

            if self.errors + self.failures:
                line_list.extend(
                    [""]
                    + self._format.error_overview(
                        self.errors, self.failures, slash_last_fail_in_overview
                    )
                )

            if self.skip_map:
                line_list.extend([""] + self._format.skips(self.skip_map))

            if self.verbosity:
                line_list.insert(0, "")

            for line in line_list:
                self.stream.writeln(line)

        def __chooseWriter(self):
            if measure_time:
                return TimeWriter
            if traditional_verbose:
                return StandardVerboseWriter
            if self.verbosity > 1:
                return ImprovedVerboseWriter
            if self.verbosity > 0:
                return DotWriter
            return Writer

    return ColorTextTestResult



from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from pcs.common.interface.dto import DataTransferObject
from pcs.common.reports.dto import ReportItemDto

from .types import (
    TaskFinishType,
    TaskKillReason,
    TaskState,
)


@dataclass(frozen=True)
class CommandOptionsDto(DataTransferObject):
    request_timeout: Optional[int] = None
    effective_username: Optional[str] = None
    effective_groups: Optional[List[str]] = None


@dataclass(frozen=True)
class CommandDto(DataTransferObject):
    command_name: str
    params: Dict[str, Any]
    options: CommandOptionsDto


@dataclass(frozen=True)
class TaskIdentDto(DataTransferObject):
    task_ident: str


@dataclass(frozen=True)
class TaskResultDto(DataTransferObject):
    task_ident: str
    command: CommandDto
    reports: List[ReportItemDto]
    state: TaskState
    task_finish_type: TaskFinishType
    kill_reason: Optional[TaskKillReason]
    result: Any



# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=too-many-branches

import os.path
import sys

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(CURRENT_DIR, "{}.d".format(sys.argv[0]))


def write_file_to_stdout(file_name):
    with open(file_name) as file:
        sys.stdout.write(file.read())


def write_local_file_to_stdout(file_name):
    write_file_to_stdout(os.path.join(DATA_DIR, file_name))


def get_arg_values(argv, name):
    values = []
    next_value = len(argv)
    for i, value in enumerate(argv):
        if value == name:
            next_value = i + 1
        elif i == next_value:
            values.append(value)
    return values


def main():
    argv = sys.argv[1:]
    if not argv:
        raise AssertionError()

    arg = argv.pop(0)

    option_file_map = {
        "--list-standards": "list_standards",
        "--list-ocf-providers": "list_ocf_providers",
    }

    if arg == "--show-metadata":
        if len(argv) != 1:
            raise AssertionError()
        arg = argv[0]
        known_agents = (
            "lsb:network",
            "ocf:heartbeat:Dummy",
            "ocf:heartbeat:IPaddr2",
            "ocf:heartbeat:Filesystem",
            "ocf:pacemaker:Dummy",
            "ocf:pacemaker:HealthCPU",
            "ocf:pacemaker:remote",
            "ocf:pacemaker:Stateful",
            "ocf:pacemaker:SystemHealth",
            "stonith:fence_apc",
            "stonith:fence_ilo",
            "stonith:fence_scsi",
            "stonith:fence_xvm",
            "systemd:test@a:b",
        )
        # known_agents_map = {item.lower()}
        if arg in known_agents:
            write_local_file_to_stdout(
                "{}_metadata.xml".format(arg.replace(":", "__"))
            )
        elif arg == "ocf:pacemaker:nonexistent":
            sys.stderr.write(
                "Metadata query for ocf:pacemaker:nonexistent failed: "
                "Input/output error\n"
            )
            raise SystemExit(5)
        elif arg == "stonith:fence_noexist":
            sys.stderr.write(
                "Agent fence_noexist not found or does not support meta-data: "
                "Invalid argument (22)\nMetadata query for "
                "stonith:fence_noexist failed: Input/output error\n"
            )
            raise SystemExit(5)
        else:
            raise AssertionError()
    elif arg in option_file_map:
        if argv:
            raise AssertionError()
        write_local_file_to_stdout(option_file_map[arg])
    elif arg == "--list-agents":
        if len(argv) != 1:
            raise AssertionError()
        arg = argv[0]
        known_providers = ("ocf:heartbeat", "ocf:pacemaker", "stonith")
        if arg in known_providers:
            write_local_file_to_stdout(
                "list_agents_{}".format(arg.replace(":", "__"))
            )
        else:
            raise AssertionError()
    elif arg == "--validate":
        if get_arg_values(argv, "--output-as")[0] != "xml":
            raise AssertionError()
        provider = get_arg_values(argv, "--provider")
        is_invalid = "fake=is_invalid=True" in argv
        output = ""
        if is_invalid:
            output = """<output source="stderr">Validation failure</output>"""
        stdout = """
            <pacemaker-result api-version="2.15" request="{cmd_str}">
              <resource-agent-action action="validate" class="{standard}" type="{agent_type}"{provider_str}>
                <overrides/>
                <agent-status code="5" message="not installed" execution_code="0" execution_message="complete" reason="environment is invalid, resource considered stopped"/>
                <command code="5">
                {output}
                </command>
              </resource-agent-action>
              <status code="5" message="Not installed">
                <errors>
                  <error>crm_resource: Error performing operation: Not installed</error>
                </errors>
              </status>
            </pacemaker-result>
        """.format(
            cmd_str=" ".join(sys.argv),
            standard=get_arg_values(argv, "--class")[0],
            agent_type=get_arg_values(argv, "--agent")[0],
            provider_str=f' provider="{provider[0]}"' if provider else "",
            output=output,
        )
        sys.stdout.write(stdout)
        if is_invalid:
            raise SystemExit(1)
    else:
        raise AssertionError()


if __name__ == "__main__":
    main()



import dataclasses

from pcs.lib.file.json import JsonExporter
from pcs.lib.interface.config import ExporterInterface

from .types import ConfigV2


class ExporterV2(ExporterInterface):
    @staticmethod
    def export(config_structure: ConfigV2) -> bytes:
        data = dataclasses.asdict(config_structure)
        data["format_version"] = 2
        return JsonExporter.export(data)



from unittest import TestCase

from pcs.lib import tools


class EnvironmentFileToDictTest(TestCase):
    def test_success(self):
        data = """
# ignored comment
IgnoredOptionNoValue
NUMBER=10
OPTION=value # this is also value
ANOTHER ONE="complex value"
; another comment

"""
        expected = {
            "NUMBER": "10",
            "OPTION": "value # this is also value",
            "ANOTHER ONE": '"complex value"',
        }
        self.assertEqual(expected, tools.environment_file_to_dict(data))

    def test_empty_string(self):
        self.assertEqual({}, tools.environment_file_to_dict(""))


class DictToEnvironmentFileTest(TestCase):
    def test_success(self):
        cfg_dict = {
            "OPTION": "value",
            "ANOTHER": "option value",
            "ANOTHER ONE": '"complex value"',
        }
        expected = """# This file has been generated by pcs.
ANOTHER=option value
ANOTHER ONE="complex value"
OPTION=value
"""
        self.assertEqual(expected, tools.dict_to_environment_file(cfg_dict))



from collections import namedtuple

# pylint: disable=import-error
import pyagentx

BaseType = namedtuple("BaseType", ["data_type", "value"])


class IntegerType(BaseType):
    def __new__(cls, value):
        return super(IntegerType, cls).__new__(
            cls, data_type=pyagentx.TYPE_INTEGER, value=value
        )


class StringType(BaseType):
    def __new__(cls, value):
        return super(StringType, cls).__new__(
            cls, data_type=pyagentx.TYPE_OCTETSTRING, value=value
        )


class Oid(
    namedtuple("OidBase", ["oid", "str_oid", "data_type", "member_list"])
):
    """
    This class represents one entity in OID tree. It is possible to define MIB
    tree model for translating string (human friendly) oid into numbered oid
    used in SNMP.

    oid int -- unique oid identificator on a given layer
    str_oid string -- string oid identificator in a given layer
    data_type BaseType -- class inherited from BaseType, data type of entity
      this class represents
    member_list list of Oid -- list of members/descendants of this entity.
      If set, this entity is threated as object identifier in MIB  and data_type
      is ignored.
    """

    def __new__(cls, oid, str_oid, data_type=None, member_list=None):
        return super(Oid, cls).__new__(
            cls, oid, str_oid, data_type, member_list
        )



import re
from typing import Optional

from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.str_tools import join_multilines
from pcs.lib import sbd
from pcs.lib.communication import qdevice as qdevice_com
from pcs.lib.communication import qdevice_net as qdevice_net_com
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.corosync import config_validators as corosync_conf_validators
from pcs.lib.corosync import live as corosync_live
from pcs.lib.corosync import (
    qdevice_client,
    qdevice_net,
)
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfFacade
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names


def get_config(lib_env: LibraryEnvironment):
    """
    Extract and return quorum configuration from corosync.conf
    lib_env LibraryEnvironment
    """
    cfg = lib_env.get_corosync_conf()
    device = None
    qd_model = cfg.get_quorum_device_model()
    if qd_model is not None:
        (
            qd_model_options,
            qd_generic_options,
            qd_heuristics_options,
        ) = cfg.get_quorum_device_settings()
        device = dict(
            model=qd_model,
            model_options=qd_model_options,
            generic_options=qd_generic_options,
            heuristics_options=qd_heuristics_options,
        )
    return {
        "options": cfg.get_quorum_options(),
        "device": device,
    }


def _check_if_atb_can_be_disabled(
    service_manager: ServiceManagerInterface,
    report_processor: reports.ReportProcessor,
    corosync_conf: CorosyncConfFacade,
    was_enabled: bool,
    force: bool = False,
) -> None:
    """
    Check whenever auto_tie_breaker can be changed without affecting SBD.
    Raises LibraryError if change of ATB will affect SBD functionality.

    service_manager --
    report_processor -- report processor
    corosync_conf -- corosync conf facade
    was_enabled -- True if ATB was enabled, False otherwise
    force -- force change
    """
    if (
        was_enabled
        and not corosync_conf.is_enabled_auto_tie_breaker()
        and sbd.is_auto_tie_breaker_needed(service_manager, corosync_conf)
    ):
        report_processor.report(
            reports.ReportItem(
                severity=reports.item.get_severity(
                    reports.codes.FORCE,
                    force,
                ),
                message=(
                    reports.messages.CorosyncQuorumAtbCannotBeDisabledDueToSbd()
                ),
            )
        )
        if report_processor.has_errors:
            raise LibraryError()


def set_options(
    lib_env: LibraryEnvironment,
    options,
    skip_offline_nodes=False,
    force=False,
):
    """
    Set corosync quorum options, distribute and reload corosync.conf if live

    LibraryEnvironment lib_env
    dict options -- quorum options
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    bool force -- force changes
    """
    cfg = lib_env.get_corosync_conf()
    if lib_env.report_processor.report_list(
        corosync_conf_validators.update_quorum_options(
            options,
            cfg.get_quorum_device_model() is not None,
            cfg.get_quorum_options(),
        )
    ).has_errors:
        raise LibraryError()
    cfg.set_quorum_options(options)
    if lib_env.is_corosync_conf_live:
        _check_if_atb_can_be_disabled(
            lib_env.service_manager,
            lib_env.report_processor,
            cfg,
            cfg.is_enabled_auto_tie_breaker(),
            force,
        )
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def status_text(lib_env):
    """
    Get quorum runtime status in plain text
    """
    try:
        return corosync_live.get_quorum_status_text(lib_env.cmd_runner())
    except corosync_live.QuorumStatusReadException as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.CorosyncQuorumGetStatusError(e.reason)
            )
        ) from e


def status_device_text(lib_env, verbose=False):
    """
    Get quorum device client runtime status in plain text
    bool verbose get more detailed output
    """
    return qdevice_client.get_status_text(lib_env.cmd_runner(), verbose)


def add_device(
    lib_env: LibraryEnvironment,
    model,
    model_options,
    generic_options,
    heuristics_options,
    force_model=False,
    force_options=False,
    skip_offline_nodes=False,
):
    # pylint: disable=too-many-locals
    """
    Add a quorum device to a cluster, distribute and reload configs if live

    string model -- quorum device model
    dict model_options -- model specific options
    dict generic_options -- generic quorum device options
    dict heuristics_options -- heuristics options
    bool force_model -- continue even if the model is not valid
    bool force_options -- continue even if options are not valid
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if cfg.get_quorum_device_model():
        raise LibraryError(
            reports.ReportItem.error(reports.messages.QdeviceAlreadyDefined())
        )

    report_processor = lib_env.report_processor
    report_processor.report_list(
        corosync_conf_validators.add_quorum_device(
            model,
            model_options,
            generic_options,
            heuristics_options,
            [node.nodeid for node in cfg.get_nodes() if node.nodeid],
            force_model=force_model,
            force_options=force_options,
        )
    )

    if lib_env.is_corosync_conf_live:
        cluster_nodes_names, report_list = get_existing_nodes_names(
            cfg,
            # Pcs is unable to communicate with nodes missing names. It cannot
            # send new corosync.conf to them. That might break the cluster.
            # Hence we error out.
            error_on_missing_name=True,
        )
        report_processor.report_list(report_list)

    if report_processor.has_errors:
        raise LibraryError()

    cfg.add_quorum_device(
        model,
        model_options,
        generic_options,
        heuristics_options,
    )
    if cfg.is_quorum_device_heuristics_enabled_with_no_exec():
        lib_env.report_processor.report(
            reports.ReportItem.warning(
                reports.messages.CorosyncQuorumHeuristicsEnabledWithNoExec()
            )
        )

    # First setup certificates for qdevice, then send corosync.conf to nodes.
    # If anything fails, nodes will not have corosync.conf with qdevice in it,
    # so there is no effect on the cluster.
    if lib_env.is_corosync_conf_live:
        target_factory = lib_env.get_node_target_factory()
        target_list = target_factory.get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
        # Do model specific configuration.
        # If the model is not known to pcs and was forced, do not configure
        # anything else than corosync.conf, as we do not know what to do
        # anyway.
        if model == "net":
            qdevice_net.set_up_client_certificates(
                lib_env.cmd_runner(),
                lib_env.report_processor,
                lib_env.communicator_factory,
                # We are sure the "host" key is there, it has been validated
                # above.
                target_factory.get_target_from_hostname(model_options["host"]),
                cfg.get_cluster_name(),
                target_list,
                skip_offline_nodes,
            )

        lib_env.report_processor.report(
            reports.ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_ENABLE, "corosync-qdevice"
                )
            )
        )
        com_cmd = qdevice_com.Enable(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # everything set up, it's safe to tell the nodes to use qdevice
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)

    # Now, when corosync.conf has been reloaded, we can start qdevice service.
    if lib_env.is_corosync_conf_live:
        lib_env.report_processor.report(
            reports.ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_START, "corosync-qdevice"
                )
            )
        )
        com_cmd_start = qdevice_com.Start(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_start.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_start)


def update_device(
    lib_env: LibraryEnvironment,
    model_options,
    generic_options,
    heuristics_options,
    force_options=False,
    skip_offline_nodes=False,
):
    """
    Change quorum device settings, distribute and reload configs if live

    dict model_options -- model specific options
    dict generic_options -- generic quorum device options
    dict heuristics_options -- heuristics options
    bool force_options -- continue even if options are not valid
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    model = cfg.get_quorum_device_model()
    if not model:
        raise LibraryError(
            reports.ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    if lib_env.report_processor.report_list(
        corosync_conf_validators.update_quorum_device(
            model,
            model_options,
            generic_options,
            heuristics_options,
            [node.nodeid for node in cfg.get_nodes() if node.nodeid],
            force_options=force_options,
        )
    ).has_errors:
        raise LibraryError()
    cfg.update_quorum_device(model_options, generic_options, heuristics_options)
    if cfg.is_quorum_device_heuristics_enabled_with_no_exec():
        lib_env.report_processor.report(
            reports.ReportItem.warning(
                reports.messages.CorosyncQuorumHeuristicsEnabledWithNoExec()
            )
        )
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def remove_device_heuristics(lib_env, skip_offline_nodes=False):
    """
    Stop using quorum device heuristics, distribute and reload configs if live

    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if not cfg.get_quorum_device_model():
        raise LibraryError(
            reports.ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    cfg.remove_quorum_device_heuristics()
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def remove_device(lib_env: LibraryEnvironment, skip_offline_nodes=False):
    """
    Stop using quorum device, distribute and reload configs if live
    skip_offline_nodes continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    model = cfg.get_quorum_device_model()
    if not model:
        raise LibraryError(
            reports.ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    cfg.remove_quorum_device()

    if lib_env.is_corosync_conf_live:
        report_processor = lib_env.report_processor
        # get nodes for communication
        cluster_nodes_names, report_list = get_existing_nodes_names(
            cfg,
            # Pcs is unable to communicate with nodes missing names. It cannot
            # send new corosync.conf to them. That might break the cluster.
            # Hence we error out.
            error_on_missing_name=True,
        )
        if report_processor.report_list(report_list).has_errors:
            raise LibraryError()
        target_list = lib_env.get_node_target_factory().get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
        # fix quorum options for SBD to work properly
        if sbd.atb_has_to_be_enabled(lib_env.service_manager, cfg):
            lib_env.report_processor.report(
                reports.ReportItem.warning(
                    reports.messages.CorosyncQuorumAtbWillBeEnabledDueToSbd()
                )
            )
            cfg.set_quorum_options({"auto_tie_breaker": "1"})

        # disable qdevice
        lib_env.report_processor.report(
            reports.ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_DISABLE, "corosync-qdevice"
                )
            )
        )
        com_cmd_disable = qdevice_com.Disable(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_disable.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_disable)
        # stop qdevice
        lib_env.report_processor.report(
            reports.ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_STOP, "corosync-qdevice"
                )
            )
        )
        com_cmd_stop = qdevice_com.Stop(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_stop.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_stop)
        # handle model specific configuration
        if model == "net":
            lib_env.report_processor.report(
                reports.ReportItem.info(
                    reports.messages.QdeviceCertificateRemovalStarted()
                )
            )
            com_cmd_client_destroy = qdevice_net_com.ClientDestroy(
                lib_env.report_processor, skip_offline_nodes
            )
            com_cmd_client_destroy.set_targets(target_list)
            run_and_raise(
                lib_env.get_node_communicator(), com_cmd_client_destroy
            )

    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def _ensure_live(lib_env: LibraryEnvironment) -> None:
    if not lib_env.is_corosync_conf_live:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.LiveEnvironmentRequired(
                    [file_type_codes.COROSYNC_CONF]
                )
            )
        )
        raise LibraryError()


def device_net_certificate_setup_local(
    lib_env: LibraryEnvironment,
    qnetd_host: str,
    cluster_name: str,
) -> None:
    """
    Set up qdevice model net certificates on the local node

    qnetd_host -- address of a qnetd host
    cluster_name -- local node cluster name
    """
    _ensure_live(lib_env)
    runner = lib_env.cmd_runner()
    qnetd_target = lib_env.get_node_target_factory().get_target_from_hostname(
        qnetd_host
    )

    # get qnetd CA certificate
    com_cmd_1 = qdevice_net_com.GetCaCert(lib_env.report_processor)
    com_cmd_1.set_targets([qnetd_target])
    qnetd_ca_cert = run_and_raise(
        lib_env.communicator_factory.get_communicator(), com_cmd_1
    )[0][1]
    # init certificate storage on local node
    qdevice_net.client_setup(runner, qnetd_ca_cert)
    # create client certificate request
    cert_request = qdevice_net.client_generate_certificate_request(
        runner, cluster_name
    )
    # sign the request on qnetd host
    com_cmd_2 = qdevice_net_com.SignCertificate(lib_env.report_processor)
    com_cmd_2.add_request(qnetd_target, cert_request, cluster_name)
    signed_certificate = run_and_raise(
        lib_env.communicator_factory.get_communicator(), com_cmd_2
    )[0][1]
    # transform the signed certificate to pk12 format which can sent to nodes
    pk12 = qdevice_net.client_cert_request_to_pk12(runner, signed_certificate)
    # store final certificate
    qdevice_net.client_import_certificate_and_key(runner, pk12)


def device_net_certificate_check_local(
    lib_env: LibraryEnvironment,
    qnetd_host: str,
    cluster_name: str,
) -> bool:
    """
    Check if the local node already has qdevice model net certificates

    qnetd_host -- address of a qnetd host
    cluster_name -- local node cluster name
    """

    def get_certificate_data(text: str) -> Optional[str]:
        match = re.search(
            r"-----BEGIN CERTIFICATE-----([a-zA-Z0-9/+=]+)-----END CERTIFICATE-----",
            "".join(text.splitlines()),
        )
        return match.group(1) if match else None

    _ensure_live(lib_env)
    runner = lib_env.cmd_runner()
    qnetd_target = lib_env.get_node_target_factory().get_target_from_hostname(
        qnetd_host
    )

    if not qdevice_net.client_initialized():
        return False

    # check if certificates are present
    list_certs_cmd = [
        settings.certutil_executable,
        "-d",
        settings.corosync_qdevice_net_client_certs_dir,
        "-L",
    ]
    stdout, stderr, retval = runner.run(list_certs_cmd)
    if retval != 0:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateReadError(
                    join_multilines([stderr, stdout])
                )
            )
        )
        raise LibraryError()
    for certname in ["QNet CA", "Cluster Cert"]:
        if not re.search(rf"^{certname}\s*\S*,.*,.*$", stdout, re.MULTILINE):
            return False

    # check cluster name
    stdout, stderr, retval = runner.run(list_certs_cmd + ["-n", "Cluster Cert"])
    if retval != 0:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateReadError(
                    join_multilines([stderr, stdout])
                )
            )
        )
        raise LibraryError()
    if not re.search(
        rf'^\s*Subject: "CN={cluster_name}"$', stdout, re.MULTILINE
    ):
        return False

    # get qnetd CA certificate from qnetd host
    com_cmd = qdevice_net_com.GetCaCert(lib_env.report_processor)
    com_cmd.set_targets([qnetd_target])
    qnetd_ca_cert_text = run_and_raise(
        lib_env.communicator_factory.get_communicator(), com_cmd
    )[0][1]
    qnetd_ca_cert = get_certificate_data(qnetd_ca_cert_text.decode())
    if qnetd_ca_cert is None:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateBadFormat()
            )
        )
        raise LibraryError()

    # get qnetd CA certificate from local node
    stdout, stderr, retval = runner.run(
        list_certs_cmd + ["-n", "QNet CA", "-a"]
    )
    if retval != 0:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateReadError(
                    join_multilines([stderr, stdout])
                )
            )
        )
        raise LibraryError()
    local_ca_cert = get_certificate_data(stdout)
    if local_ca_cert is None:
        lib_env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateBadFormat()
            )
        )
        raise LibraryError()

    return qnetd_ca_cert == local_ca_cert


def set_expected_votes_live(lib_env, expected_votes):
    """
    set expected votes in live cluster to specified value
    numeric expected_votes desired value of expected votes
    """
    try:
        votes_int = int(expected_votes)
        if votes_int < 1:
            raise ValueError()
    except ValueError:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "expected votes", expected_votes, "positive integer"
                )
            )
        ) from None

    corosync_live.set_expected_votes(lib_env.cmd_runner(), votes_int)



from unittest import (
    TestCase,
    mock,
)

from pcs import status
from pcs.cli.common.errors import CmdLineInputError

from pcs_test.tools.misc import dict_to_modifiers


@mock.patch("pcs.status.print")
class XmlStatus(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["status"])
        self.lib.status = mock.Mock(spec_set=["pacemaker_status_xml"])

    def _call_cmd(self, argv=None):
        status.xml_status(self.lib, argv or [], dict_to_modifiers({}))

    def test_success(self, mock_print):
        xml_status = """
            <xml>
                <pcmk status data />
            </xml>
        """
        self.lib.status.pacemaker_status_xml.return_value = xml_status
        self._call_cmd([])
        self.lib.status.pacemaker_status_xml.assert_called_once_with()
        mock_print.assert_called_once_with(xml_status.strip())

    def test_argv(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["x"])
        self.assertIsNone(cm.exception.message)
        mock_print.assert_not_called()



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation


def fixture_resources_xml(resources_xml_list):
    return """
        <resources>
            {0}
        </resources>
    """.format(
        "\n".join(resources_xml_list)
    )


def fixture_primitive_xml(primitive_id):
    return f"""
        <primitive class="ocf" id="{primitive_id}" provider="heartbeat"
            type="Dummy"
        >
            <operations>
                <op id="{primitive_id}-monitor-interval-10s" interval="10s"
                    name="monitor" timeout="20s"/>
            </operations>
        </primitive>
    """


def fixture_group_xml(group_id, primitive_xml_list):
    return """
        <group id="{group_id}">
            {group_resources}
        </group>
    """.format(
        group_id=group_id,
        group_resources="\n".join(primitive_xml_list),
    )


def fixture_clone_xml(clone_id, clone_resource_xml):
    return """
        <clone id="{clone_id}">
            {clone_resource}
        </clone>
    """.format(
        clone_id=clone_id,
        clone_resource=clone_resource_xml,
    )


FIXTURE_AGROUP_XML = fixture_group_xml(
    "AGroup",
    [
        fixture_primitive_xml("A1"),
        fixture_primitive_xml("A2"),
        fixture_primitive_xml("A3"),
    ],
)


FIXTURE_CONSTRAINTS_CONFIG_XML = """
    <constraints>
        <rsc_location id="location-AGroup-rh7-1-INFINITY" node="rh7-1"
            rsc="AGroup" score="INFINITY"/>
        <rsc_location id="location-TagGroupOnly-rh7-1-INFINITY"
            node="rh7-1" rsc="TagGroupOnly" score="INFINITY"/>
    </constraints>
"""

FIXTURE_CLONE_TAG_CONSTRAINTS = """
    <constraints>
        <rsc_location id="location-AGroup-rh7-1-INFINITY" node="rh7-1"
            rsc="AGroup-clone" score="INFINITY"
        />
        <rsc_location id="location-TagGroupOnly-rh7-1-INFINITY"
            node="rh7-1" rsc="TagGroupOnly" score="INFINITY"
        />
    </constraints>
"""


FIXTURE_CLONE_CONSTRAINT = """
    <constraints>
        <rsc_location id="location-AGroup-rh7-1-INFINITY" node="rh7-1"
            rsc="AGroup-clone" score="INFINITY"
        />
    </constraints>
"""


FIXTURE_TAGS_CONFIG_XML = """
    <tags>
        <tag id="TagGroupOnly">
            <obj_ref id="AGroup"/>
        </tag>
        <tag id="TagNotGroupOnly">
            <obj_ref id="AGroup"/>
            <obj_ref id="A1"/>
            <obj_ref id="A2"/>
            <obj_ref id="A3"/>
        </tag>
    </tags>
"""


FIXTURE_TAGS_RESULT_XML = """
    <tags>
        <tag id="TagNotGroupOnly">
            <obj_ref id="A1"/>
            <obj_ref id="A2"/>
            <obj_ref id="A3"/>
        </tag>
    </tags>
"""


class TestGroupMixin:
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_cib_resource_group_ungroup")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        xml_manip = XmlManipulation.from_file(self.empty_cib)
        xml_manip.append_to_first_tag_name("resources", FIXTURE_AGROUP_XML)
        xml_manip.append_to_first_tag_name(
            "configuration",
            FIXTURE_TAGS_CONFIG_XML,
        )
        xml_manip.append_to_first_tag_name(
            "constraints",
            """
            <rsc_location id="location-AGroup-rh7-1-INFINITY" node="rh7-1"
                rsc="AGroup" score="INFINITY"/>
            """,
            """
            <rsc_location id="location-TagGroupOnly-rh7-1-INFINITY"
                node="rh7-1" rsc="TagGroupOnly" score="INFINITY"/>
            """,
        )
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()


class GroupDeleteRemoveUngroupBase(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
    TestGroupMixin,
):
    command = None

    def assert_tags_xml(self, expected_xml):
        self.assert_resources_xml_in_cib(
            expected_xml,
            get_cib_part_func=lambda cib: etree.tostring(
                etree.parse(cib).findall(".//tags")[0],
            ),
        )

    def assert_constraint_xml(self, expected_xml):
        self.assert_resources_xml_in_cib(
            expected_xml,
            get_cib_part_func=lambda cib: etree.tostring(
                etree.parse(cib).findall(".//constraints")[0],
            ),
        )

    def test_nonexistent_group(self):
        self.assert_pcs_fail(
            ["resource"] + self.command + ["NonExistentGroup"],
            "Error: Group 'NonExistentGroup' does not exist\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml([FIXTURE_AGROUP_XML]),
        )
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CONSTRAINTS_CONFIG_XML)

    def test_not_a_group_id(self):
        self.assert_pcs_fail(
            ["resource"] + self.command + ["A1"],
            "Error: Group 'A1' does not exist\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml([FIXTURE_AGROUP_XML]),
        )
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CONSTRAINTS_CONFIG_XML)

    def test_whole_group(self):
        self.assert_effect(
            ["resource"] + self.command + ["AGroup"],
            fixture_resources_xml(
                [
                    fixture_primitive_xml("A1"),
                    fixture_primitive_xml("A2"),
                    fixture_primitive_xml("A3"),
                ],
            ),
            stderr_full=(
                "Removing Constraint - location-TagGroupOnly-rh7-1-INFINITY\n"
                "Removing Constraint - location-AGroup-rh7-1-INFINITY\n"
            ),
        )
        self.assert_tags_xml(FIXTURE_TAGS_RESULT_XML)
        self.assert_constraint_xml("<constraints/>")

    def test_specified_resources(self):
        self.assert_effect(
            ["resource"] + self.command + ["AGroup", "A1", "A3"],
            fixture_resources_xml(
                [
                    fixture_group_xml(
                        "AGroup",
                        [
                            fixture_primitive_xml("A2"),
                        ],
                    ),
                    fixture_primitive_xml("A1"),
                    fixture_primitive_xml("A3"),
                ],
            ),
        )
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CONSTRAINTS_CONFIG_XML)

    def test_all_resources(self):
        self.assert_effect(
            ["resource"] + self.command + ["AGroup", "A1", "A2", "A3"],
            fixture_resources_xml(
                [
                    fixture_primitive_xml("A1"),
                    fixture_primitive_xml("A2"),
                    fixture_primitive_xml("A3"),
                ],
            ),
            stderr_full=(
                "Removing Constraint - location-TagGroupOnly-rh7-1-INFINITY\n"
                "Removing Constraint - location-AGroup-rh7-1-INFINITY\n"
            ),
        )
        self.assert_tags_xml(FIXTURE_TAGS_RESULT_XML)
        self.assert_constraint_xml("<constraints/>")

    def test_cloned_group(self):
        self.assert_pcs_success("resource clone AGroup".split())
        self.assert_pcs_fail(
            ["resource"] + self.command + ["AGroup"],
            "Error: Cannot remove all resources from a cloned group\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(
                [fixture_clone_xml("AGroup-clone", FIXTURE_AGROUP_XML)],
            )
        )
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CLONE_TAG_CONSTRAINTS)

    def test_cloned_group_all_resources_specified(self):
        self.assert_pcs_success("resource clone AGroup".split())
        self.assert_pcs_fail(
            ["resource"] + self.command + ["AGroup", "A1", "A2", "A3"],
            "Error: Cannot remove all resources from a cloned group\n",
        )
        self.assert_resources_xml_in_cib(
            fixture_resources_xml(
                [fixture_clone_xml("AGroup-clone", FIXTURE_AGROUP_XML)],
            )
        )
        self.assert_tags_xml(FIXTURE_TAGS_CONFIG_XML)
        self.assert_constraint_xml(FIXTURE_CLONE_TAG_CONSTRAINTS)

    def test_cloned_group_with_one_resource(self):
        self.assert_pcs_success("resource clone AGroup".split())
        self.assert_pcs_success("resource ungroup AGroup A1 A2".split())
        self.assert_effect(
            ["resource"] + self.command + ["AGroup"],
            fixture_resources_xml(
                [
                    fixture_clone_xml(
                        "AGroup-clone",
                        fixture_primitive_xml("A3"),
                    ),
                    fixture_primitive_xml("A1"),
                    fixture_primitive_xml("A2"),
                ],
            ),
            stderr_full="Removing Constraint - location-TagGroupOnly-rh7-1-INFINITY\n",
        )
        self.assert_tags_xml(FIXTURE_TAGS_RESULT_XML)
        self.assert_constraint_xml(FIXTURE_CLONE_CONSTRAINT)


class ResourceUngroup(GroupDeleteRemoveUngroupBase, TestCase):
    command = ["ungroup"]


class GroupDelete(GroupDeleteRemoveUngroupBase, TestCase):
    command = ["group", "delete"]


class GroupRemove(GroupDeleteRemoveUngroupBase, TestCase):
    command = ["group", "remove"]



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

# pylint: disable=invalid-name
# pylint: disable=too-many-public-methods

empty_cib = rc("cib-empty.xml")


class ACLTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_acl")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def testEnableDisable(self):
        self.assert_pcs_success("acl disable".split())
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )
        self.assert_pcs_success("acl enable".split())
        self.assert_pcs_success(
            ["acl"],
            "ACLs are enabled\n\n",
        )
        self.assert_pcs_success("acl disable".split())
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )

    def testUserGroupCreateDeleteWithRoles(self):
        self.assert_pcs_success(
            "acl role create role1 read xpath /xpath1/ write xpath /xpath2/".split()
        )
        self.assert_pcs_success(
            "acl role create role2 deny xpath /xpath3/ deny xpath /xpath4/".split()
        )
        self.assert_pcs_success(
            "acl role create role3 read xpath /xpath5/ read xpath /xpath6/".split()
        )
        self.assert_pcs_fail(
            "acl user create user1 roleX".split(),
            "Error: ACL role 'roleX' does not exist\n",
        )
        self.assert_pcs_fail(
            "acl user create user1 role1 roleX".split(),
            "Error: ACL role 'roleX' does not exist\n",
        )
        self.assert_pcs_fail(
            "acl group create group1 roleX".split(),
            "Error: ACL role 'roleX' does not exist\n",
        )
        self.assert_pcs_fail(
            "acl group create group1 role1 roleX".split(),
            "Error: ACL role 'roleX' does not exist\n",
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: read xpath /xpath6/ (role3-read-1)
                """
            ),
        )
        self.assert_pcs_success("acl user create user1 role1 role2".split())
        self.assert_pcs_success("acl group create group1 role1 role3".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role1 role2
                Group: group1
                  Roles: role1 role3
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: read xpath /xpath6/ (role3-read-1)
                """
            ),
        )
        self.assert_pcs_fail(
            "acl role create group1".split(),
            "Error: 'group1' already exists\n",
        )
        self.assert_pcs_fail(
            "acl role create role1".split(),
            "Error: 'role1' already exists\n",
        )
        self.assert_pcs_fail(
            "acl user create user1".split(),
            "Error: 'user1' already exists\n",
        )
        self.assert_pcs_fail(
            "acl group create group1".split(),
            "Error: 'group1' already exists\n",
        )
        self.assert_pcs_fail(
            "acl group create role1".split(),
            "Error: 'role1' already exists\n",
        )
        self.assert_pcs_fail(
            "acl role assign role1 to noexist".split(),
            "Error: ACL group/ACL user 'noexist' does not exist\n",
        )
        self.assert_pcs_fail(
            "acl role assign noexist to user1".split(),
            "Error: ACL role 'noexist' does not exist\n",
        )
        self.assert_pcs_success("acl role assign role3 to user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role1 role2 role3
                Group: group1
                  Roles: role1 role3
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: read xpath /xpath6/ (role3-read-1)
                """
            ),
        )
        self.assert_pcs_fail(
            "acl role unassign noexist from user1".split(),
            "Error: Role 'noexist' is not assigned to 'user1'\n",
        )
        self.assert_pcs_fail(
            "acl role unassign role3 from noexist".split(),
            "Error: ACL group/ACL user 'noexist' does not exist\n",
        )
        self.assert_pcs_success("acl role unassign role3 from user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role1 role2
                Group: group1
                  Roles: role1 role3
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: read xpath /xpath6/ (role3-read-1)
                """
            ),
        )
        self.assert_pcs_success("acl role unassign role2 from user1".split())
        self.assert_pcs_success("acl role unassign role1 from user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                Group: group1
                  Roles: role1 role3
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: read xpath /xpath6/ (role3-read-1)
                """
            ),
        )
        self.assert_pcs_success("acl role delete role3".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success("acl role assign role2 to user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role2
                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success("acl role assign role1 user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role2 role1
                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success(
            "acl role unassign role2 from user1 --autodelete".split()
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role1
                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success(
            "acl role unassign role1 from user1 --autodelete".split()
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success("acl user create user1 role1 role2".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role1 role2
                Group: group1
                  Roles: role1
                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success("acl role delete role1 --autodelete".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles: role2
                Role: role2
                  Permission: deny xpath /xpath3/ (role2-deny)
                  Permission: deny xpath /xpath4/ (role2-deny-1)
                """
            ),
        )
        self.assert_pcs_success("acl role delete role2".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                """
            ),
        )

    def testUserGroupCreateDelete(self):
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_success("acl user create user2".split())
        self.assert_pcs_fail(
            "acl user create user1".split(),
            "Error: 'user1' already exists\n",
        )
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_success("acl group create group2".split())
        self.assert_pcs_fail(
            "acl group create group1".split(),
            "Error: 'group1' already exists\n",
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                User: user2
                  Roles:
                Group: group1
                  Roles:
                Group: group2
                  Roles:
                """
            ),
        )
        self.assert_pcs_fail(
            "acl group delete user1".split(),
            "Error: ACL group 'user1' does not exist\n",
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                User: user2
                  Roles:
                Group: group1
                  Roles:
                Group: group2
                  Roles:
                """
            ),
        )
        self.assert_pcs_success("acl group delete group2".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                User: user2
                  Roles:
                Group: group1
                  Roles:
                """
            ),
        )
        self.assert_pcs_success("acl group remove group1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user1
                  Roles:
                User: user2
                  Roles:
                """
            ),
        )
        self.assert_pcs_success("acl user delete user1".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                User: user2
                  Roles:
                """
            ),
        )
        self.assert_pcs_success("acl user remove user2".split())
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )

    def testRoleCreateDelete(self):
        self.assert_pcs_fail(
            "acl role create role0 read".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 read //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 read xpath".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 read id".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 readX xpath //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 read xpathX //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test read".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test read //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test read xpath".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test read id".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test readX xpath //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 description=test read xpathX //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test read".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test read //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test read xpath".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test read id".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test readX xpath //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_fail(
            "acl role create role0 desc=test read xpathX //resources".split(),
            stderr_start="\nUsage: pcs acl role create...",
        )
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )
        self.assert_pcs_success("acl role create role0".split())
        self.assert_pcs_fail(
            "acl role create role0".split(), "Error: 'role0' already exists\n"
        )
        self.assert_pcs_success(
            ["acl", "role", "create", "role0d", "description=empty role"]
        )
        self.assert_pcs_success(
            "acl role create role1 read xpath /xpath/".split()
        )
        self.assert_pcs_success(
            [
                "acl",
                "role",
                "create",
                "role2",
                "description=with description",
                "READ",
                "XPATH",
                "/xpath/",
            ]
        )
        self.assert_pcs_success(
            (
                "acl role create role3 Read XPath /xpath_query/ wRiTe xpATH "
                "/xpath_query2/ deny xpath /xpath_query3/"
            ).split()
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role0
                Role: role0d
                  Description: empty role
                Role: role1
                  Permission: read xpath /xpath/ (role1-read)
                Role: role2
                  Description: with description
                  Permission: read xpath /xpath/ (role2-read)
                Role: role3
                  Permission: read xpath /xpath_query/ (role3-read)
                  Permission: write xpath /xpath_query2/ (role3-write)
                  Permission: deny xpath /xpath_query3/ (role3-deny)
                """
            ),
        )
        self.assert_pcs_success("acl role delete role2".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role0
                Role: role0d
                  Description: empty role
                Role: role1
                  Permission: read xpath /xpath/ (role1-read)
                Role: role3
                  Permission: read xpath /xpath_query/ (role3-read)
                  Permission: write xpath /xpath_query2/ (role3-write)
                  Permission: deny xpath /xpath_query3/ (role3-deny)
                """
            ),
        )
        self.assert_pcs_fail(
            "acl role delete role2".split(),
            "Error: ACL role 'role2' does not exist\n",
        )
        self.assert_pcs_success("acl role delete role1".split())
        self.assert_pcs_success("acl role remove role3".split())
        self.assert_pcs_success("acl role remove role0".split())
        self.assert_pcs_success("acl role remove role0d".split())
        self.assert_pcs_success(
            ["acl"],
            "ACLs are disabled, run 'pcs acl enable' to enable\n\n",
        )

    def testPermissionAddDelete(self):
        self.assert_pcs_success(
            "acl role create role1 read xpath /xpath1/ write xpath /xpath2/".split()
        )
        self.assert_pcs_success(
            "acl role create role2 read xpath /xpath3/ write xpath /xpath4/".split()
        )
        self.assert_pcs_success(
            "acl role create role3 read xpath /xpath5/ write xpath /xpath6/".split()
        )
        self.assert_pcs_success(
            "acl config".split(),
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                Role: role2
                  Permission: read xpath /xpath3/ (role2-read)
                  Permission: write xpath /xpath4/ (role2-write)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: write xpath /xpath6/ (role3-write)
                """
            ),
        )
        self.assert_pcs_success(
            "acl permission add role1 deny xpath /myxpath1/".split()
        )
        self.assert_pcs_success(
            "acl permission add role4 deny xpath /myxpath2/".split()
        )
        self.assert_pcs_success(
            "acl config".split(),
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                  Permission: deny xpath /myxpath1/ (role1-deny)
                Role: role2
                  Permission: read xpath /xpath3/ (role2-read)
                  Permission: write xpath /xpath4/ (role2-write)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: write xpath /xpath6/ (role3-write)
                Role: role4
                  Permission: deny xpath /myxpath2/ (role4-deny)
                """
            ),
        )
        self.assert_pcs_success("acl permission delete role4-deny".split())
        self.assert_pcs_fail(
            "acl permission delete role4-deny".split(),
            "Error: ACL permission 'role4-deny' does not exist\n",
        )
        self.assert_pcs_fail(
            "acl permission remove role4-deny".split(),
            "Error: ACL permission 'role4-deny' does not exist\n",
        )
        self.assert_pcs_success(
            "acl config".split(),
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                  Permission: deny xpath /myxpath1/ (role1-deny)
                Role: role2
                  Permission: read xpath /xpath3/ (role2-read)
                  Permission: write xpath /xpath4/ (role2-write)
                Role: role3
                  Permission: read xpath /xpath5/ (role3-read)
                  Permission: write xpath /xpath6/ (role3-write)
                Role: role4
                """
            ),
        )
        self.assert_pcs_success("acl permission delete role3-read".split())
        self.assert_pcs_success("acl permission delete role3-write".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                  Permission: read xpath /xpath1/ (role1-read)
                  Permission: write xpath /xpath2/ (role1-write)
                  Permission: deny xpath /myxpath1/ (role1-deny)
                Role: role2
                  Permission: read xpath /xpath3/ (role2-read)
                  Permission: write xpath /xpath4/ (role2-write)
                Role: role3
                Role: role4
                """
            ),
        )
        self.assert_pcs_success("acl permission remove role1-read".split())
        self.assert_pcs_success("acl permission remove role1-write".split())
        self.assert_pcs_success("acl permission remove role1-deny".split())
        self.assert_pcs_success("acl permission remove role2-read".split())
        self.assert_pcs_success("acl permission remove role2-write".split())
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                Role: role2
                Role: role3
                Role: role4
                """
            ),
        )
        self.assert_pcs_fail(
            "acl permission add role1 read".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read xpath".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 readX xpath //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read xpathX //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read xpath".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read id".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy readX xpath //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read xpathX //resources".split(),
            stderr_start="\nUsage: pcs acl permission add...",
        )
        self.assert_pcs_success(
            ["acl"],
            dedent(
                """\
                ACLs are disabled, run 'pcs acl enable' to enable

                Role: role1
                Role: role2
                Role: role3
                Role: role4
                """
            ),
        )

    def test_can_add_permission_for_existing_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_success(
            "acl permission add role1 read id role2".split()
        )

    def test_can_add_permission_for_existing_xpath(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success(
            "acl permission add role1 read xpath //nodes".split()
        )

    def test_can_not_add_permission_for_nonexisting_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_fail(
            "acl permission add role1 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_add_permission_for_nonexisting_id_in_later_part(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_fail(
            "acl permission add role1 read id role2 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_add_permission_for_nonexisting_role_with_bad_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_fail(
            "acl permission add #bad-name read id role1".split(),
            "Error: invalid ACL role '#bad-name'"
            + ", '#' is not a valid first character for a ACL role\n",
        )

    def test_can_create_role_with_permission_for_existing_id(self):
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_success("acl role create role1 read id role2".split())

    def test_can_not_crate_role_with_permission_for_nonexisting_id(self):
        self.assert_pcs_fail(
            "acl role create role1 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_create_role_with_bad_name(self):
        self.assert_pcs_fail(
            "acl role create #bad-name".split(),
            "Error: invalid ACL role '#bad-name'"
            + ", '#' is not a valid first character for a ACL role\n",
        )

    def test_fail_on_unknown_role_method(self):
        self.assert_pcs_fail(
            "acl role unknown whatever".split(),
            stderr_start="\nUsage: pcs acl role ...",
        )

    def test_assign_unassign_role_to_user(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_success("acl role assign role1 user user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 user user1".split(),
            "Error: Role 'role1' is already assigned to 'user1'\n",
        )
        self.assert_pcs_success("acl role unassign role1 user user1".split())
        self.assert_pcs_fail(
            "acl role unassign role1 user user1".split(),
            "Error: Role 'role1' is not assigned to 'user1'\n",
        )

    def test_assign_unassign_role_to_user_not_existing_user(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to user group1".split(),
            "Error: 'group1' is not an ACL user\n",
        )

    def test_assign_unassign_role_to_user_with_to(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_success("acl role assign role1 to user user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to user user1".split(),
            "Error: Role 'role1' is already assigned to 'user1'\n",
        )
        self.assert_pcs_success(
            "acl role unassign role1 from user user1".split()
        )
        self.assert_pcs_fail(
            "acl role unassign role1 from user user1".split(),
            "Error: Role 'role1' is not assigned to 'user1'\n",
        )

    def test_assign_unassign_role_to_group(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_success("acl role assign role1 group group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 group group1".split(),
            "Error: Role 'role1' is already assigned to 'group1'\n",
        )
        self.assert_pcs_success("acl role unassign role1 group group1".split())
        self.assert_pcs_fail(
            "acl role unassign role1 group group1".split(),
            "Error: Role 'role1' is not assigned to 'group1'\n",
        )

    def test_assign_unassign_role_to_group_not_existing_group(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to group user1".split(),
            "Error: ACL group 'user1' does not exist\n",
        )

    def test_assign_unassign_role_to_group_with_to(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_success("acl role assign role1 to group group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to group group1".split(),
            "Error: Role 'role1' is already assigned to 'group1'\n",
        )
        self.assert_pcs_success(
            "acl role unassign role1 from group group1".split()
        )
        self.assert_pcs_fail(
            "acl role unassign role1 from group group1".split(),
            "Error: Role 'role1' is not assigned to 'group1'\n",
        )



from functools import partial

from pcs import (
    node,
    usage,
)
from pcs.cli.common.routing import create_router

node_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.node(argv)),
        "maintenance": partial(node.node_maintenance_cmd, enable=True),
        "unmaintenance": partial(node.node_maintenance_cmd, enable=False),
        "standby": partial(node.node_standby_cmd, enable=True),
        "unstandby": partial(node.node_standby_cmd, enable=False),
        "attribute": node.node_attribute_cmd,
        "utilization": node.node_utilization_cmd,
        # pcs-to-pcsd use only
        "pacemaker-status": node.node_pacemaker_status,
    },
    ["node"],
)



from unittest import TestCase

from pcs.common import reports
from pcs.common.reports import codes as report_codes
from pcs.lib.commands.sbd import disable_sbd

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class DisableSbd(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync-3nodes.conf"
        self.node_list = ["rh7-1", "rh7-2", "rh7-3"]
        self.config.env.set_known_nodes(self.node_list)

    def test_success(self):
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[[dict(label=node)] for node in self.node_list],
        )
        self.config.http.sbd.disable_sbd(node_labels=self.node_list)
        disable_sbd(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.SERVICE_ACTION_STARTED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    instance="",
                )
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    node=node,
                    instance="",
                )
                for node in self.node_list
            ]
            + [
                fixture.warn(
                    report_codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES
                )
            ]
        )

    def test_some_node_names_missing(self):
        self.corosync_conf_name = "corosync-some-node-names.conf"
        self.node_list = ["rh7-2"]

        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[[dict(label=node)] for node in self.node_list],
        )
        self.config.http.sbd.disable_sbd(node_labels=self.node_list)

        disable_sbd(self.env_assist.get_env())

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=False,
                ),
                fixture.info(
                    reports.codes.SERVICE_ACTION_STARTED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    instance="",
                ),
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    node=node,
                    instance="",
                )
                for node in self.node_list
            ]
            + [
                fixture.warn(
                    report_codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES
                )
            ]
        )

    def test_all_node_names_missing(self):
        self.config.corosync_conf.load(filename="corosync-no-node-names.conf")
        self.env_assist.assert_raise_library_error(
            lambda: disable_sbd(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=False,
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_NO_NODES_DEFINED,
                ),
            ]
        )

    def test_node_offline(self):
        err_msg = "Failed connect to rh7-3:2224; No route to host"
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(
            communication_list=[
                {"label": "rh7-1"},
                {"label": "rh7-2"},
                {
                    "label": "rh7-3",
                    "was_connected": False,
                    "errno": 7,
                    "error_msg": err_msg,
                },
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: disable_sbd(self.env_assist.get_env()),
            [],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    node="rh7-3",
                    reason=err_msg,
                    command="remote/check_auth",
                )
            ]
        )

    def test_success_node_offline_skip_offline(self):
        err_msg = "Failed connect to rh7-3:2224; No route to host"
        online_nodes_list = ["rh7-2", "rh7-3"]
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(
            communication_list=[
                {
                    "label": "rh7-1",
                    "was_connected": False,
                    "errno": 7,
                    "error_msg": err_msg,
                },
                {"label": "rh7-2"},
                {"label": "rh7-3"},
            ]
        )
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[
                [dict(label=node)] for node in self.node_list[1:]
            ],
        )
        self.config.http.sbd.disable_sbd(node_labels=online_nodes_list)
        disable_sbd(self.env_assist.get_env(), ignore_offline_nodes=True)
        self.env_assist.assert_reports(
            [fixture.warn(report_codes.OMITTING_NODE, node="rh7-1")]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_STARTED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    instance="",
                )
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    node=node,
                    instance="",
                )
                for node in online_nodes_list
            ]
            + [
                fixture.warn(
                    report_codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES
                )
            ]
        )

    def test_set_stonith_watchdog_timeout_fails_on_some_nodes(self):
        err_msg = "Error"
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[
                [
                    {
                        "label": "rh7-1",
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": err_msg,
                    }
                ],
                [
                    {
                        "label": "rh7-2",
                        "response_code": 400,
                        "output": "FAILED",
                    }
                ],
                [{"label": "rh7-3"}],
            ]
        )
        self.config.http.sbd.disable_sbd(node_labels=self.node_list)
        disable_sbd(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="rh7-1",
                    reason=err_msg,
                    command="remote/set_stonith_watchdog_timeout_to_zero",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="rh7-2",
                    reason="FAILED",
                    command="remote/set_stonith_watchdog_timeout_to_zero",
                ),
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_STARTED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    instance="",
                )
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    node=node,
                    instance="",
                )
                for node in self.node_list
            ]
            + [
                fixture.warn(
                    report_codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES
                )
            ]
        )

    def test_set_stonith_watchdog_timeout_fails_on_all_nodes(self):
        err_msg = "Error"
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[
                [dict(label=node, response_code=400, output=err_msg)]
                for node in self.node_list
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: disable_sbd(self.env_assist.get_env()),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=node,
                    reason=err_msg,
                    command="remote/set_stonith_watchdog_timeout_to_zero",
                )
                for node in self.node_list
            ]
            + [
                fixture.error(
                    report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE,
                )
            ]
        )

    def test_disable_failed(self):
        err_msg = "Error"
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.pcmk.set_stonith_watchdog_timeout_to_zero(
            communication_list=[[dict(label=node)] for node in self.node_list],
        )
        self.config.http.sbd.disable_sbd(
            communication_list=[
                {"label": "rh7-1"},
                {"label": "rh7-2"},
                {"label": "rh7-3", "response_code": 400, "output": err_msg},
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: disable_sbd(self.env_assist.get_env()),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.SERVICE_ACTION_STARTED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    instance="",
                )
            ]
            + [
                fixture.info(
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    action=reports.const.SERVICE_ACTION_DISABLE,
                    service="sbd",
                    node=node,
                    instance="",
                )
                for node in self.node_list[:2]
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="rh7-3",
                    reason=err_msg,
                    command="remote/sbd_disable",
                )
            ]
        )



from typing import (
    Iterable,
    Mapping,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.tools import timeout_to_seconds
from pcs.common.types import StringSequence
from pcs.lib import (
    sbd,
    validate,
)
from pcs.lib.cib import nvpair_multi
from pcs.lib.cib.tools import (
    IdProvider,
    get_crm_config,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.errors import LibraryError
from pcs.lib.resource_agent import ResourceAgentFacade

_READONLY_CLUSTER_PROPERTY_LIST = [
    "cluster-infrastructure",
    "cluster-name",
    "dc-version",
    "have-watchdog",
    "last-lrm-refresh",
]
_DEFAULT_CLUSTER_PROPERTY_SET_ID = "cib-bootstrap-options"


def _validate_stonith_watchdog_timeout_property(
    service_manager: ServiceManagerInterface,
    value: str,
    force: bool = False,
) -> reports.ReportItemList:
    report_list: reports.ReportItemList = []
    original_value = value
    # if value is not empty, try to convert time interval string
    if value:
        seconds = timeout_to_seconds(value)
        if seconds is None:
            # returns empty list because this should be reported by
            # ValueTimeInterval validator
            return report_list
        value = str(seconds)
    if sbd.is_sbd_enabled(service_manager):
        report_list.extend(
            sbd.validate_stonith_watchdog_timeout(
                validate.ValuePair(original_value, value), force
            )
        )
    else:
        if value not in ["", "0"]:
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.StonithWatchdogTimeoutCannotBeSet(
                        reports.const.SBD_NOT_SET_UP
                    ),
                )
            )
    return report_list


def validate_set_cluster_properties(
    cluster_property_facade_list: Iterable[ResourceAgentFacade],
    properties_set_id: str,
    configured_properties: StringSequence,
    new_properties: Mapping[str, str],
    service_manager: ServiceManagerInterface,
    force: bool = False,
) -> reports.ReportItemList:
    """
    Validate that cluster properties and their values can be set.

    cluster_property_facade_list -- facades for cluster properties metadata
    properties_set_id -- id of the properties set to be updated
    configured_properties -- names of currently configured cluster properties
    new_properties -- dictionary of properties and their values to be set
    service_manager -- manager for system daemon services
    force -- if True, produce warnings instead of errors
    """
    # pylint: disable=too-many-locals
    possible_properties_dict = {
        parameter.name: parameter
        for facade in cluster_property_facade_list
        for parameter in facade.metadata.parameters
        if parameter.name not in _READONLY_CLUSTER_PROPERTY_LIST
    }
    severity = reports.get_severity(reports.codes.FORCE, force)

    to_be_set_properties = {}
    to_be_removed_properties = []
    for name, value in new_properties.items():
        if value != "":
            to_be_set_properties[name] = value
        else:
            to_be_removed_properties.append(name)

    report_list = validate.validate_set_unset_items(
        to_be_set_properties.keys(),
        to_be_removed_properties,
        configured_properties,
        reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
        reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
        properties_set_id,
        severity=severity,
    )

    report_list.extend(
        validate.NamesIn(
            possible_properties_dict.keys(),
            option_type="cluster property",
            banned_name_list=_READONLY_CLUSTER_PROPERTY_LIST,
            severity=severity,
        ).validate(
            # Allow removing properties unknown to pacemaker while preventing
            # setting them. Prevent removing read-only properties.
            {
                name: value
                for name, value in new_properties.items()
                if not (
                    value == ""
                    and name not in _READONLY_CLUSTER_PROPERTY_LIST
                    and name not in possible_properties_dict
                )
            }
        )
    )

    validators: list[validate.ValidatorInterface] = []
    for property_name in to_be_set_properties:
        if property_name not in possible_properties_dict:
            # unknow properties are reported by NamesIn validator
            continue
        property_metadata = possible_properties_dict[property_name]
        if property_metadata.type == "boolean":
            validators.append(
                validate.ValuePcmkBoolean(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "integer":
            validators.append(
                validate.ValuePcmkInteger(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "percentage":
            validators.append(
                validate.ValuePcmkPercentage(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "select":
            validators.append(
                validate.ValueIn(
                    property_metadata.name,
                    property_metadata.enum_values or [],
                    severity=severity,
                )
            )
        elif property_metadata.type == "time":
            # make stonith-watchdog-timeout value not forcable
            validators.append(
                validate.ValueTimeInterval(
                    property_metadata.name,
                    severity=severity
                    if property_metadata.name != "stonith-watchdog-timeout"
                    else reports.ReportItemSeverity.error(),
                )
            )
    report_list.extend(
        validate.ValidatorAll(validators).validate(to_be_set_properties)
    )

    # Only validate SWT if it is being set, or if it is being removed and it
    # actually exists in the current configuration.
    if "stonith-watchdog-timeout" in new_properties and (
        new_properties["stonith-watchdog-timeout"]
        or "stonith-watchdog-timeout" in configured_properties
    ):
        report_list.extend(
            _validate_stonith_watchdog_timeout_property(
                service_manager,
                new_properties["stonith-watchdog-timeout"],
                force=force,
            )
        )

    return report_list


def get_cluster_property_set_element_legacy(
    cib: _Element, id_provider: IdProvider
) -> _Element:
    """
    Return the first cluster_property_set element. If the element does not
    exist, try to create cluster_property_set element with id of value
    'cib-bootstrap-options'. Raise an error in case the id is already used.

    cib -- cib tree
    id_provider -- checks id uniqueness and books ids if set
    """
    crm_config_el = get_crm_config(cib)
    property_el_list = nvpair_multi.find_nvsets(
        crm_config_el, nvpair_multi.NVSET_PROPERTY
    )
    if property_el_list:
        return property_el_list[0]
    if id_provider.book_ids(_DEFAULT_CLUSTER_PROPERTY_SET_ID):
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.CannotCreateDefaultClusterPropertySet(
                    _DEFAULT_CLUSTER_PROPERTY_SET_ID
                )
            )
        )
    return nvpair_multi.nvset_append_new(
        crm_config_el,
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        nvpair_multi.NVSET_PROPERTY,
        {},
        {"id": _DEFAULT_CLUSTER_PROPERTY_SET_ID},
        nvset_rule=None,
    )



from dataclasses import dataclass
from typing import (
    Collection,
    Sequence,
)

from pcs.common.permissions.types import (
    PermissionAccessType,
    PermissionTargetType,
)
from pcs.common.types import StringSequence


@dataclass(frozen=True)
class ClusterEntry:
    name: str
    nodes: StringSequence


@dataclass(frozen=True)
class PermissionEntry:
    name: str
    type: PermissionTargetType
    allow: Collection[PermissionAccessType]


@dataclass(frozen=True)
class ClusterPermissions:
    local_cluster: Sequence[PermissionEntry]


@dataclass(frozen=True)
class ConfigV2:
    data_version: int
    clusters: Sequence[ClusterEntry]
    permissions: ClusterPermissions
